#include "FuncionesSSE.h"
#include <cstring>
#include <cmath>
#include <stdio.h>
#include <stdlib.h>

#define forn(i,n) for(int i = 0 ; i < (int) (n); i++)

extern "C" void sse_invertir_internal(unsigned char *Imagen, int h, int w, int ws);
extern "C" void sse_bordesSimples_internal(unsigned char *Imagen, int h, int w, int ws);
extern "C" void sse_bordesSobel_internal(unsigned char *Imagen, int h, int w, int ws, 
                                        int *movx, int *movy, int *valorH, int *valorV, char *out);
extern "C" void sse_BlackAndWhite_internal(unsigned char *Imagen, int h, int w, int ws);
extern "C" void sse_zoom_internal(unsigned char *Imagen, int h, int w, int ws, 
                                    int inih, int iniw, int finh, int finw,char *out);
extern "C" void sse_fotoBinaria_internal(unsigned char *Imagen, int h, int w, int ws, int corte);
extern "C" void sse_filtroGauseano_internal(unsigned char *Imagen, int h, int w, int ws, bool ByN, int *movx, int *movy, int *valores);
extern "C" void sse_bordesCanny_internal(unsigned char *Imagen, int h, int w, int ws, bool ByN, bool binario,
                                        int *movx, int *movy, int *valorH, int *valorV, char *Emag, char *Egra);

extern "C" void sse_calcular_corte_fotoBinaria_internal(unsigned char *Imagen, int h, int w, int ws, 
                                                int corteint, int *sumtecho, int *sumpiso, int *cantecho, int *canpiso);



extern "C" void sse_zoomBillinear_internal(unsigned char *Imagen, int h, int w, int ws);
extern "C" void sse_duplicarContraste_internal(unsigned char *Imagen, int h, int w, int ws);
extern "C" void sse_colorBalance_internal(unsigned char *Imagen, int h, int w, int ws, unsigned RED, unsigned GREEN, unsigned BLUE);
extern "C" void sse_bokeh_internal(unsigned char *Imagen, int h, int w, int ws, unsigned kernelSize, unsigned AREA, unsigned *kernel, unsigned char *aux);
extern "C" void sse_nearestRot_internal(unsigned char *Imagen, int h, int w, int ws, unsigned char * aux, float sinAlpha, float cosAlpha);

typedef unsigned char uchar;

void sse_bordesCanny(unsigned char *Imagen, int h, int w, int ws, bool ByN, bool binario)
{
	if (!ByN)
		sse_BlackAndWhite(Imagen,h,w,ws);

	int movx[9] =	{-3, -3,-3,0,  0, 0, 3,  3, 3};
	int movy[9] =	{ 0,-ws,ws,0,-ws,ws, 0,-ws,ws};
	int valorH[9] = {-2, -1,-1,0,  0, 0, 2,  1, 1};
	int valorV[9] = { 0, -1, 1,0, -2, 2, 0, -1, 1};

	char *Emag = (char*)malloc(h*ws);
	char *Egra = (char*)malloc(h*ws);
	
    sse_bordesCanny_internal(Imagen, h, w, ws, ByN, binario, movx, movy, valorH, valorV, Emag, Egra);
    
	free(Emag);
	free(Egra);
	w/=3;
	h++;
	if (binario)
		sse_fotoBinaria(Imagen,h,w,ws);
}



/*listo*/
int sse_calcular_corte_fotoBinaria(unsigned char *Imagen, int h, int w, int ws)
{
	double corte = 0;
	double cortenuevo = 150;
	w*=3;
	while ((corte - cortenuevo) > 1 || (corte - cortenuevo) < -1 )
	{
		corte = cortenuevo;
		int sumtecho = 0;
		int sumpiso = 0;
		int cantecho = 0;
		int canpiso = 0;
		int corteint = corte;
		
        sse_calcular_corte_fotoBinaria_internal(Imagen,h,w,ws,corteint, &sumtecho, &sumpiso, &cantecho, &canpiso);
        
		
		cortenuevo = ((double)sumpiso / (double)canpiso + (double)sumtecho / (double)cantecho)/2;
	}
	

	return (int)corte;
}


void sse_fotoBinaria(unsigned char *Imagen, int h, int w, int ws)
{
	
	int corte = sse_calcular_corte_fotoBinaria(Imagen, h,w,ws);
    sse_fotoBinaria_internal(Imagen,h,w,ws,corte);
}

void sse_filtroGauseano(unsigned char *Imagen, int h, int w, int ws, bool ByN)
{
	int movx[25] = {-3,-3,-3,0,0,0,3,3,3,-6,-6,-6,-6,-6,6,6,6,6,6,-3,-3,3,3,0,0};
	int movy[25] = {0,-1*ws,1*ws,0,-1*ws,1*ws,0,-1*ws,1*ws,-2*ws,-1*ws,0,1*ws,2*ws,-2*ws,-1*ws,0,1*ws,2*ws,-2*ws,2*ws,-2*ws,2*ws,-2*ws,2*ws};
	int valores[25] = {26,16,16,41,26,26,26,16,16,1,4,7,4,1,1,4,7,4,1,4,4,4,4,7,7};
	
    sse_filtroGauseano_internal(Imagen, h, w, ws, ByN, movx, movy, valores);
	
}

void sse_zoom(unsigned char *Imagen, int h, int w, int ws)
{
	char *out = (char*)malloc(h*ws);
	int inih =  h/2-h/4;
	int iniw =  (w/2-w/4)*3;
	int finh =  h/2+h/4;
	int finw =  (w/2+w/4 -1)*3 ;
	w*=3;
	sse_zoom_internal(Imagen, h, w, ws, inih, iniw, finh, finw,out);
	free(out);
}
void sse_bordesSobel(unsigned char *Imagen, int h, int w, int ws)
{
	int movx[9] =   {-3,-3,-3,0, 0,0, 3, 3,3};
	int movy[9] =  { 0,-ws, ws,0,-ws,ws, 0,-ws,ws};
	int valorH[9] = {-2,-1,-1,0, 0,0, 2, 1,1};
	int valorV[9] = { 0,-1, 1,0,-2,2, 0,-1,1};
	char *out = (char*)malloc(h*ws);
    
    sse_bordesSobel_internal(Imagen, h, w, ws, movx, movy, valorH, valorV, out);
	
	free(out);
}
void sse_invertir(unsigned char *Imagen, int h, int w, int ws)
{
	sse_invertir_internal(Imagen, h, w, ws);
}

void sse_BlackAndWhite(unsigned char *Imagen, int h, int w, int ws)
{
	sse_BlackAndWhite_internal(Imagen,h,w,ws);
}


void sse_bordesSimples(unsigned char *Imagen, int h, int w, int ws)
{
	sse_bordesSimples_internal(Imagen,h,w,ws);
}






void sse_zoomBillinear(unsigned char *Imagen, int h, int w, int ws)
{
	sse_zoom(Imagen, h, w, ws);
    sse_zoomBillinear_internal(Imagen,h,w,ws);
}

void sse_duplicarContraste(unsigned char *Imagen, int h, int w, int ws)
{
    sse_duplicarContraste_internal(Imagen,h,w,ws);
}

void sse_colorBalance(unsigned char *Imagen, int h, int w, int ws, unsigned RED,unsigned GREEN,unsigned BLUE)
{
    sse_colorBalance_internal(Imagen,h,w,ws, RED, GREEN, BLUE);
}

void sse_bokeh(unsigned char *Imagen, int h, int w, int ws, unsigned kernelSize)
{
    unsigned char *aux = (unsigned char*)malloc(ws*h);
    memcpy(aux, Imagen, ws*h);
    unsigned *kernel = (unsigned *)malloc(kernelSize * sizeof(unsigned));
    unsigned AREA = 0;
    forn(i,kernelSize)
        AREA += 2 * (kernel[i] = sqrt(kernelSize * kernelSize - i*i)) - 1;
    AREA = AREA * 2 - (2 * kernel[0] - 1);
    
    sse_bokeh_internal(Imagen, h, w, ws, kernelSize, AREA, kernel, aux);
    
    free(kernel);
    free(aux);
}

void sse_nearestRot(unsigned char *Imagen, int h, int w, int ws, float alpha)
{
    const float sinAlpha = sin(alpha);
    const float cosAlpha = cos(alpha);
    uchar *aux = (uchar*)malloc(ws*h);
    memcpy(aux, Imagen, ws*h);    
    sse_nearestRot_internal(Imagen,h,w,ws, aux, sinAlpha, cosAlpha);
    free(aux);
}
