#include "FuncionesCUDA.h"
#include <cmath>

typedef unsigned char uchar;
#define forn(i,n) for(int i = 0 ; i < (int) (n); i++)
#define forsn(i,s,n) for(int i = (s) ; i < (int) (n); i++)


__global__ void d_zoomBillinear(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws)
{	
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	//int k = theadIdx.y;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
	{
		int origen =  (w/2-w/4)*3+((h/2-h/4) + ((blockDim.x * blockIdx.x + threadIdx.x)/2)) * ws;
		
		for(int i= 0 ; i< w; i++)
		{
			out[y + i*3 ] = I[origen + (i/2)*3 ];
			out[y + i*3 +1] = I[origen + ((i/2)*3)+1];
			out[y + i*3 +2] = I[origen + ((i/2)*3)+2];
		}
	}
}	


void cu_zoomBillinear(unsigned char *Imagen, int h, int w, int ws)
{
	unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_zoomBillinear<<<blocks,threads>>>(d_I,d_out,h, w,ws);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
}

__global__ void d_duplicarContraste(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws, unsigned c1, unsigned c2, unsigned c3)
{	
    int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	//int k = theadIdx.y;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
	{	
		int newCol = 0;
		for(int i = 0 ; i < w ; i ++)
		{
			newCol = 2 * I[y + i * 3] - c1;
			if (newCol < 0) newCol = 0;
			if (newCol > 255) newCol = 255;
			out[y +  i * 3] = newCol;
			
			newCol = 2 * I[y + i * 3 + 1] - c2;
			if (newCol < 0) newCol = 0;
			if (newCol > 255) newCol = 255;
			out[y +  i * 3 + 1] = newCol;
			
			newCol = 2 * I[y + i * 3 + 2] - c3;
			if (newCol < 0) newCol = 0;
			if (newCol > 255) newCol = 255;
			out[y +  i * 3 + 2] = newCol;            
		}	
	}
}

void cu_duplicarContraste(unsigned char *Imagen, int h, int w, int ws)
{
	uchar *base = Imagen;
	unsigned xm[3] = {0,0,0};
    for(int i= 0 ; i< h ; i++)
	{
        const uchar *pos = base;
		for(int j = 0 ; j < w ; j ++)
		{
            forn(col,3)
                xm[col] += *pos++;
		}
        base += ws;
	}
    forn(col,3)
        xm[col] /= h*w;

	unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_duplicarContraste<<<blocks,threads>>>(d_I,d_out,h, w,ws,xm[0],xm[1],xm[2]);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
}

__global__ void d_colorBalance(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws, unsigned c1, unsigned c2, unsigned c3)
{	
    int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	//int k = theadIdx.y;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
	{
		int newCol = 0;
		for(int i = 0 ; i < w ; i ++)
		{
                newCol = (c1 * I[y+i*3] ) / 128;
                if (newCol > 255) newCol = 255;
                out[y+i*3] = newCol;
                
                newCol = (c2 * I[y+i*3+1] ) / 128;
                if (newCol > 255) newCol = 255;
                out[y+i*3+1] = newCol;
                
                newCol = (c3 * I[y+i*3+2] ) / 128;
                if (newCol > 255) newCol = 255;
                out[y+i*3+2] = newCol;
            
		}
	}
}
void cu_colorBalance(unsigned char *Imagen, int h, int w, int ws, unsigned c1, unsigned c2, unsigned c3)
{
	unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_colorBalance<<<blocks,threads>>>(d_I,d_out,h, w,ws,c1,c2,c3);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
}
/*
	uchar *aux = (uchar*)malloc(WS*h);
    memcpy(aux, frame->imageData, WS*h);
    unsigned *k = (unsigned *)malloc(kernelSize * sizeof(unsigned));
    unsigned AREA = 0;
    forn(i,kernelSize)
        AREA += 2 * (k[i] = sqrt(kernelSize * kernelSize - i*i)) - 1;
    AREA = AREA * 2 - (2 * k[0] - 1);
    
    uchar *output = (uchar*)frame->imageData;
    uchar *input = aux;
    for(int i= 0 ; i< h ; i++)
	{
        unsigned total[3] = {0,0,0};
        forsn(ip, max(-int(kernelSize)+1, -i), min(h-i,int(kernelSize)))
        {
            forn(j,k[abs(ip)])
            if (j < w)
            forn(col,3)
                total[col] += input[WS*ip + 3 * j + col];
        }
        
        // En este punto ya estan computados los totales iniciales.
        
        uchar *inputPos = input;
        uchar *outputPos = output;
		for(int j = 0 ; j < w ; j ++)
		{
            forn(col,3)
            {
                *outputPos++ = total[col] / AREA;
                // Quitar uno a izquierda, meter uno a derecha

                forsn(ip, max(-int(kernelSize)+1, -i), min(h-i,int(kernelSize)))
                {
                    const int D = k[abs(ip)];
                    if (j-D+1 >= 0)
                        total[col] -=  inputPos[WS*ip - 3*(D-1)];
                    if (j+D < w)
                        total[col] += inputPos[WS*ip + 3*D];
                }
                inputPos++;
            }
		}
        output += WS;
        input  += WS;
	}
    free(k);
    free(aux);
/**/
__global__ void d_bokeh(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws, unsigned kernelSize, unsigned *k, unsigned AREA)
{	
	int i = (blockDim.x * blockIdx.x + threadIdx.x);
    int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	//int k = theadIdx.y;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
	{
		unsigned total[3] = {0,0,0};
        forsn(ip, max(-int(kernelSize)+1, -i), min(h-i,int(kernelSize)))
        {
            forn(j,k[abs(ip)])
            if (j < w)
            forn(col,3)
                total[col] += I[y + ws*ip + 3 * j + col];
        }
        
        // En este punto ya estan computados los totales iniciales.
        
		for(int j = 0 ; j < w ; j ++)
		{
            forn(col,3)
            {
                out[y + j *3 + col] = total[col] / AREA;
                // Quitar uno a izquierda, meter uno a derecha

                forsn(ip, max(-int(kernelSize)+1, -i), min(h-i,int(kernelSize)))
                {
                    const int D = k[abs(ip)];
                    if (j-D+1 >= 0)
                        total[col] -=  I[y+j*3 + col +ws*ip - 3*(D-1)];
                    if (j+D < w)
                        total[col] += I[y+j*3 + col + ws*ip + 3*D];
                }
            }
		}
	}
}

void cu_bokeh(unsigned char *Imagen, int h, int w, int ws, unsigned kernelSize)
{
	unsigned *k = (unsigned *)malloc(kernelSize * sizeof(unsigned));
	unsigned AREA = 0;
    forn(i,kernelSize)
        AREA += 2 * (k[i] = sqrt(kernelSize * kernelSize - i*i)) - 1;
    AREA = AREA * 2 - (2 * k[0] - 1);
	unsigned char *d_I, *d_out;
	unsigned *d_k;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMalloc(&d_k, kernelSize * sizeof(unsigned));
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	cudaMemcpy(d_k,k, kernelSize * sizeof(unsigned), cudaMemcpyHostToDevice);
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_bokeh<<<blocks,threads>>>(d_I,d_out,h, w,ws,kernelSize, d_k, AREA);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
	cudaFree(d_k);
}
/*void nearestRot(IplImage *frame, float alpha)
{    
	
    uchar *aux = (uchar*)malloc(WS*h);
    memcpy(aux, frame->imageData, WS*h);    
    uchar *output = (uchar*)frame->imageData;
    const int cx = h/2, cy = w/2;
    for(int i= 0 ; i< h ; i++)
	{
        uchar *outputPos = output;
		for(int j = 0 ; j < w ; j ++)
		{
            const int ri = round(cx + cosAlpha * (i - cx) - sinAlpha * (j - cy) );
            const int rj = round(cy + sinAlpha * (i - cx) + cosAlpha * (j - cy) );
            if (0 <= ri && ri < h && 0 <= rj && rj < w)
                forn(col,3) *outputPos++ = aux[WS*ri+3*rj+col];
            else
                forn(col,3) *outputPos++ = 0;
		}
        output += WS;
	}
    free(aux);
}*/
__global__ void d_nearestRot(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws, float alpha, float sinAlpha, float cosAlpha)
{	
	int i = (blockDim.x * blockIdx.x + threadIdx.x);
    int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	//int k = theadIdx.y;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
	{
		const int cx = h/2, cy = w/2;
		for(int j = 0 ; j < w ; j ++)
		{
            const int ri = round(cx + cosAlpha * (i - cx) - sinAlpha * (j - cy) );
            const int rj = round(cy + sinAlpha * (i - cx) + cosAlpha * (j - cy) );
            if (0 <= ri && ri < h && 0 <= rj && rj < w)
                forn(col,3) out[y+j*3+col] = I[ws*ri+3*rj+col];
            else
                forn(col,3) out[y+j*3+col] = 0;
		}
	}
}
void cu_nearestRot(unsigned char *Imagen, int h, int w, int ws, float alpha)
{
	const float sinAlpha = sin(alpha);
    const float cosAlpha = cos(alpha);
    unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_nearestRot<<<blocks,threads>>>(d_I,d_out,h, w,ws,alpha,sinAlpha,cosAlpha);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
}

/*lentos*/
__global__ void d_filtroGauseano_bw(const unsigned char *I,unsigned char *out,const int *movx,const int *movy,const int *valores,const int h, const int w,const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-2 && blockDim.x * blockIdx.x + threadIdx.x  > 1)
	{
		for(int i= 2 ; i< w-2 ; i++)
		{
			int sum = 0;
			for(int k = 0 ; k < 25; k++)
			{
				sum+= (I[y+movx[k]*ws+(i+movy[k])*3])*valores[k] ;				
			}
			sum /= 273;
			
			out[y+i*3+0] = sum;
			out[y+i*3+1] = sum;
			out[y+i*3+2] = sum;
		}		
	
	}
}
__global__ void d_filtroGauseano(const unsigned char *I,unsigned char *out,const int *movx,const int *movy,const int *valores,const int h, const int w,const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-2 && blockDim.x * blockIdx.x + threadIdx.x  > 1)
	{
		for(int i= 2 ; i< (w-2)*3 ; i++)
		{
			int sum = 0;
			for(int k = 0 ; k < 25; k++)
			{
				sum+= (I[y+movx[k]*ws+(i)+movy[k]*3])*valores[k] ;				
			}
			sum /= 273;
			
			out[y+i+0] = sum;
		}		
	
	}
}


void cu_filtroGauseano(unsigned char *Imagen, int h, int w, int ws, bool ByN)
{
	int movx[25] = {-1,-1,-1,0,0,0,1,1,1,-2,-2,-2,-2,-2,2,2,2,2,2,-1,-1,1,1,0,0};
	int movy[25] = {0,-1,1,0,-1,1,0,-1,1,-2,-1,0,1,2,-2,-1,0,1,2,-2,2,-2,2,-2,2};
	int valores[25] = {26,16,16,41,26,26,26,16,16,1,4,7,4,1,1,4,7,4,1,4,4,4,4,7,7};
	int *d_movx, *d_movy, *d_valores;

	unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMalloc(&d_movx,25*sizeof(int));
	cudaMalloc(&d_movy,25*sizeof(int));
	cudaMalloc(&d_valores,25*sizeof(int));

	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	cudaMemcpy(d_out,Imagen, size, cudaMemcpyHostToDevice);//*//
	cudaMemcpy(d_movx,movx, 25*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(d_movy,movy, 25*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(d_valores,valores, 25*sizeof(int), cudaMemcpyHostToDevice);

	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	if (ByN)
		d_filtroGauseano_bw<<<blocks,threads>>>(d_I,d_out,d_movx,d_movy,d_valores,h, w,ws);
	else 
		d_filtroGauseano<<<blocks,threads>>>(d_I,d_out,d_movx,d_movy,d_valores,h, w,ws);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
	cudaFree(d_movx);
	cudaFree(d_movy);
	cudaFree(d_valores);
}

/*LISTOOSSSS*/
__global__ void d_bordesCanny_magYgrad(const unsigned char *I,unsigned char *mag, unsigned char *grad ,const int *movx,const int *movy,const int *valorH,const int *valorV,const int h, const int w,const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1 && blockDim.x * blockIdx.x + threadIdx.x  > 0)
	{
		for(int i= 1 ; i< w-1 ; i++)
		{
			int valH = 0;
			int valV = 0;
			for(int k = 0; k < 9;k++)
			{
				valH+= ((int)I[y+(movy[k])*ws+(i+movx[k])*3]) * valorH[k];
				valV+= ((int)I[y+(movy[k])*ws+(i+movx[k])*3]) * valorV[k];
			}
			int val = abs(valH) + abs(valV) ;
			if (val > 255) val = 255;
			mag[y+i*3]= val;
			unsigned char aux = 90;
			if (valV != 0)
			{
				//float vald = (float)valH/(float)valV;
				//aux = 90;
				if (valH*1000 > valV*-425 && valH*1000 < valV * 425)
					aux = 0;
				if (valH*1000 <valV*-425 && valH*1000 < valV *-2425)
					aux = 135;
				if (valH*1000 >valV*425 && valH*1000 < valV *2425)
					aux = 45;
			}
			grad[y+i*3]= aux;
		}
	}
}
__global__ void d_bordesCanny(unsigned char *I,const unsigned char *mag,const unsigned char *grad ,const int *movx,const int *movy,const int *valorH,const int *valorV,const int h, const int w,const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1 && blockDim.x * blockIdx.x + threadIdx.x  > 0)
	{
		for(int i= 1 ; i< w-1 ; i++)
		{
			int k = 0;
			if (grad[y+i*3] == 0)
				k = 4;
			if (grad[y+i*3] == 135) // estan bien??
				k = 8;
			if (grad[y+i*3] == 45) // estan bien??
				k = 7;
			if (grad[y+i*3] == 90)
				k = 0;
			if (mag[y+i*3] < mag[y+ movy[k] * ws + (i+movx[k])*3] || mag[y+i*3] < mag[y+ movy[k]*-1 * ws + (i+movx[k]*-1)*3])
			{
				I[y+i*3] = 0;	
				I[y+i*3+1] = 0;	
				I[y+i*3+2] = 0;	
			}
			else
			{
				I[y+i*3] = mag[y+i*3];	
				I[y+i*3+1] = mag[y+i*3];	
				I[y+i*3+2] = mag[y+i*3];	
			}
		}
	}
}

void cu_bordesCanny(unsigned char *Imagen, int h, int w, int ws, bool ByN, bool binario)
{
	if (!ByN)
		cu_BlackAndWhite(Imagen,h,w,ws);


	int movx[9] =   {-1,-1,-1,0, 0,0,1, 1,1};
	int movy[9] =   {0, -1, 1,0,-1,1,0,-1,1};
	int valorH[9] = {-2,-1,-1,0, 0,0, 2,1,1};
	int valorV[9] = {0, -1, 1,0,-2,2, 0,-1,1};
	int *d_movx, *d_movy, *d_valorH, *d_valorV;

	unsigned char *d_I, *d_grad, *d_mag ;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_grad,size);
	cudaMalloc(&d_mag,size);
	cudaMalloc(&d_movx,9*sizeof(int));
	cudaMalloc(&d_movy,9*sizeof(int));
	cudaMalloc(&d_valorH,9*sizeof(int));
	cudaMalloc(&d_valorV,9*sizeof(int));

	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	cudaMemcpy(d_movx,movx, 9*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(d_movy,movy, 9*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(d_valorH,valorH, 9*sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(d_valorV,valorV, 9*sizeof(int), cudaMemcpyHostToDevice);

	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	
	d_bordesCanny_magYgrad<<<blocks,threads>>>(d_I,d_mag,d_grad,d_movx,d_movy,d_valorH,d_valorV,h, w,ws);
	
	d_bordesCanny<<<blocks,threads>>>(d_I,d_mag,d_grad,d_movx,d_movy,d_valorH,d_valorV,h, w,ws);
	

	cudaMemcpy(Imagen,d_I, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_mag);
	cudaFree(d_grad);
	cudaFree(d_movx);
	cudaFree(d_movy);
	cudaFree(d_valorV);
	cudaFree(d_valorH);



	if (binario)
		cu_fotoBinaria(Imagen,h,w,ws);
}

__global__ void d_fotoBinaria_transformacion(unsigned char *I, const int corte, const int h, const int w, const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
	{
		for(int i= 0 ; i< w ; i++)
		{				
			int suma = 0;
			suma += (unsigned char)I[y+i*3];
			suma += (unsigned char)I[y+i*3+1];
			suma += (unsigned char)I[y+i*3+2];			
			if (suma > corte)
			{
				I[y+i*3] = 255;
				I[y+i*3+1] = 255;
				I[y+i*3+2] = 255;
			}
			else
			{				
				I[y+i*3] = 0;
				I[y+i*3+1] = 0;
				I[y+i*3+2] = 0;
			}
				
		}
	}
}
__global__ void d_fotoBinaria_calcular(const unsigned char *I, int *out, const int corte, const int h, const int w, const int ws)
{
	int j =  (blockDim.x * blockIdx.x + threadIdx.x);
	int y = j * ws;
	if (j < h)
	{
		out[j*4] = 0; out[j*4+1] = 0;out[j*4+2] = 0;out[j*4+3] = 0;
		for(int i= 0 ; i< w ; i++)
		{
			int aux = 0;
			aux += I[y+i*3];
			aux += I[y+i*3+1];
			aux += I[y+i*3+2];
			if (corte < aux)
			{
				out[j*4 + 0]++;
				out[j*4 + 1]+= aux;			
			}
			else
			{
				out[j*4 + 2]++;
				out[j*4 + 3]+= aux;							
			}
		}		
	}
}

void cu_fotoBinaria(unsigned char *Imagen, int h, int w, int ws)
{
	unsigned char *d_I;
	int *d_out;
	int *h_out;
	h_out = (int*)malloc(h*4*sizeof(int));
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,h*4*sizeof(int));
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;

	double corte  = 0;	
	double cortenuevo = 150;
	while (abs(corte - cortenuevo) > 1)
	{
		corte = cortenuevo;
		int sumtecho = 0;
		int sumpiso = 0;
		int cantecho = 0;
		int canpiso = 0;
		
		d_fotoBinaria_calcular<<<blocks,threads>>>(d_I,d_out,corte,h, w,ws);
		cudaMemcpy(h_out,d_out, h*4*sizeof(int), cudaMemcpyDeviceToHost);
		for(int j = 0; j < h; j++)
		{
			cantecho+= h_out[j*4 + 0];
			sumtecho+= h_out[j*4 + 1];
			canpiso += h_out[j*4 + 2];
			sumpiso += h_out[j*4 + 3];
		}
		cortenuevo = ((double)sumpiso / (double)canpiso + (double)sumtecho / (double)cantecho)/2;
	}

	d_fotoBinaria_transformacion<<<blocks,threads>>>(d_I,(int)corte,h, w,ws);
	cudaMemcpy(Imagen,d_I, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
	free(h_out);
}




__global__ void d_zoom(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws)
{	
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	//int k = theadIdx.y;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
	{
		int origen =  (w/2-w/4)*3+((h/2-h/4) + ((blockDim.x * blockIdx.x + threadIdx.x)/2)) * ws;
		
		for(int i= 0 ; i< w; i++)
		{
			out[y + i*3 ] = I[origen + (i/2)*3 ];
			out[y + i*3 +1] = I[origen + ((i/2)*3)+1];
			out[y + i*3 +2] = I[origen + ((i/2)*3)+2];
		}
	}
}	


void cu_zoom(unsigned char *Imagen, int h, int w, int ws)
{
	unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	cudaMemcpy(d_out,Imagen, size, cudaMemcpyHostToDevice);//*//
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_zoom<<<blocks,threads>>>(d_I,d_out,h, w,ws);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
}

__global__ void d_bordesSobel(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws)
{	
	
	//__shared__ int movx[9];// =   {-1,-1,-1,0, 0,0, 1, 1,1};
	//__shared__ int movy[9];// =   { 0,-1, 1,0,-1,1, 0,-1,1};
	//__shared__ int valorH[9];// = {-2,-1,-1,0, 0,0, 2, 1,1};
	//__shared__ int valorV[9];// = { 0,-1, 1,0,-2,2, 0,-1,1};
	//__syncthreads();
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	//int k = theadIdx.y;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1 && blockDim.x * blockIdx.x + threadIdx.x  != 0)
		for(int i= 3 ; i< (w -1)*3; i++)
		{
			int valH = 0;
			int valV = 0;
			valH+= (I[y+0*ws+(i-3)]) * -2;
			valV+= (I[y+0*ws+(i-3)]) * 0;

			valH+= (I[y+1*ws+(i-3)]) * -1;
			valV+= (I[y+1*ws+(i-3)]) * 1;
						
			valH+= (I[y-1*ws+(i-3)]) * -1;
			valV+= (I[y-1*ws+(i-3)]) * -1;
												
			valH+= (I[y+1*ws+(i)]) * 0;
			valV+= (I[y+1*ws+(i)]) * 2;
						
			valH+= (I[y-1*ws+(i)]) * 0;
			valV+= (I[y-1*ws+(i)]) * -2;
						
			valH+= (I[y+0*ws+(i+3)]) * 2;
			valV+= (I[y+0*ws+(i+3)]) * 0;
						
			valH+= (I[y-1*ws+(i+3)]) * 1;
			valV+= (I[y-1*ws+(i+3)]) * -1;

			valH+= (I[y+1*ws+(i+3)]) * 1;
			valV+= (I[y+1*ws+(i+3)]) * 1;
			int val = abs(valH) + abs(valV) ;
			if (val > 255) val = 255;
			out[y + i]= val;
		}
	//
}


void cu_bordesSobel(unsigned char *Imagen, int h, int w, int ws)
{
	unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	cudaMemcpy(d_out,Imagen, size, cudaMemcpyHostToDevice);//*//
	//dim3 threads (32,9);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_bordesSobel<<<blocks,threads>>>(d_I,d_out,h, w,ws);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
}


__global__ void d_bordesSimples(const unsigned char *I, unsigned char *out, const int h, const int w, const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h-1)
		for(int i= 1 ; i< (w -1)*3; i+=3)
		{
			int sum  = 0;
			//derecha
			sum += abs(I[y+i] - I[y+i+3]);
			sum += abs(I[y+i+1] -I[y+i+3+1]);
			sum += abs(I[y+i+2] - I[y+i+3+2]);
			//abajo
			sum += abs(I[y+i] - I[y+ws+i]);
			sum += abs(I[y+i+1] -I[y+ws+i+1]);
			sum += abs(I[y+i+2] - I[y+ws+i+2]);
			
			unsigned char val = sum;
			if (sum > 255)
				val = 255;
			out[y+i+2] = val;
			out[y+i+1] = val;
			out[y+i] = val;

		}
}


void cu_bordesSimples(unsigned char *Imagen, int h, int w, int ws)
{
	unsigned char *d_I, *d_out;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMalloc(&d_out,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	cudaMemcpy(d_out,Imagen, size, cudaMemcpyHostToDevice);//*//
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_bordesSimples<<<blocks,threads>>>(d_I,d_out,h, w,ws);
	cudaMemcpy(Imagen,d_out, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
	cudaFree(d_out);
}

__global__ void d_invertir(unsigned char *I, const int h, const int w, const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h)
		for(int i= 0 ; i< w*3; i++)
		{
			I[y+i] = 255-I[y+i];		
		}
}

void cu_invertir(unsigned char *Imagen, int h, int w, int ws)
{
	unsigned char *d_I;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_invertir<<<blocks,threads>>>(d_I,h, w,ws);
	cudaMemcpy(Imagen,d_I, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
}

__global__ void d_BlackAndWhite(unsigned char *I, const int h, const int w, const int ws)
{
	int y = (blockDim.x * blockIdx.x + threadIdx.x) * ws;
	if (blockDim.x * blockIdx.x + threadIdx.x < h)
		for(int i= 0 ; i< w*3; i+=3)
		{			
			int sum  = 0;
			sum += I[y+i] * 30;
			sum += I[y+i+1]   * 59;
			sum += I[y+i+2]   * 11;
			sum/=100;
			I[y+i]  = (sum);
			I[y+i+1] = (sum);
			I[y+i+2] = (sum);	
		}
	// Rojo 30%, Verde 59%, Azul 11%
}
void cu_BlackAndWhite(unsigned char *Imagen, int h, int w, int ws)
{
	unsigned char *d_I;
	int size = ws*h;
	cudaMalloc(&d_I,size);
	cudaMemcpy(d_I,Imagen, size, cudaMemcpyHostToDevice);
	int threads = 256;
	int blocks = (h+threads-1)/threads;
	d_BlackAndWhite<<<blocks,threads>>>(d_I,h, w,ws);
	cudaMemcpy(Imagen,d_I, size, cudaMemcpyDeviceToHost);
	cudaFree(d_I);
}

/*
// Variables
float* h_A;
float* h_B;
float* h_C;
float* d_A;
float* d_B;
float* d_C;
bool noprompt = false;

// Functions
void CleanupResources(void);
void RandomInit(float*, int);
void ParseArguments(int, char**);

// Device code
__global__ void VecAdd(const float* A, const float* B, float* C, int N)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < N)
        C[i] = A[i] + B[i];
}

// Host code
int main(int argc, char** argv)
{
    shrQAStart(argc, argv);

    printf("Vector Addition\n");
    int N = 50000;
    size_t size = N * sizeof(float);
    ParseArguments(argc, argv);

    // Allocate input vectors h_A and h_B in host memory
    h_A = (float*)malloc(size);
    if (h_A == 0) CleanupResources();
    h_B = (float*)malloc(size);
    if (h_B == 0) CleanupResources();
    h_C = (float*)malloc(size);
    if (h_C == 0) CleanupResources();
    
    // Initialize input vectors
    RandomInit(h_A, N);
    RandomInit(h_B, N);

    // Allocate vectors in device memory
    cutilSafeCall( cudaMalloc((void**)&d_A, size) );
    cutilSafeCall( cudaMalloc((void**)&d_B, size) );
    cutilSafeCall( cudaMalloc((void**)&d_C, size) );

    // Copy vectors from host memory to device memory
    cutilSafeCall( cudaMemcpy(d_A, h_A, size, cudaMemcpyHostToDevice) );
    cutilSafeCall( cudaMemcpy(d_B, h_B, size, cudaMemcpyHostToDevice) );

    // Invoke kernel
    int threadsPerBlock = 256;
    int blocksPerGrid = (N + threadsPerBlock - 1) / threadsPerBlock;
    VecAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A, d_B, d_C, N);
    cutilCheckMsg("kernel launch failure");
#ifdef _DEBUG
    cutilSafeCall( cutilDeviceSynchronize() );
#endif

    // Copy result from device memory to host memory
    // h_C contains the result in host memory
    cutilSafeCall( cudaMemcpy(h_C, d_C, size, cudaMemcpyDeviceToHost) );
    
    // Verify result
    int i;
    for (i = 0; i < N; ++i) {
        float sum = h_A[i] + h_B[i];
        if (fabs(h_C[i] - sum) > 1e-5)
            break;
    }

    CleanupResources();
    shrQAFinishExit(argc, (const char **)argv, (i==N) ? QA_PASSED : QA_FAILED);
}

void CleanupResources(void)
{
    // Free device memory
    if (d_A)
        cudaFree(d_A);
    if (d_B)
        cudaFree(d_B);
    if (d_C)
        cudaFree(d_C);

    // Free host memory
    if (h_A)
        free(h_A);
    if (h_B)
        free(h_B);
    if (h_C)
        free(h_C);
        
    cutilDeviceReset();
}

// Allocates an array with random float entries.
void RandomInit(float* data, int n)
{
    for (int i = 0; i < n; ++i)
        data[i] = rand() / (float)RAND_MAX;
}

// Parse program arguments
void ParseArguments(int argc, char** argv)
{
    for (int i = 0; i < argc; ++i)
        if (strcmp(argv[i], "--noprompt") == 0 ||
			strcmp(argv[i], "-noprompt") == 0) 
		{
            noprompt = true;
            break;
        }
}*/
